import random
import os
from pathlib import Path

FOLDER = Path.home()
FILE = os.path.join(FOLDER, 'maze.txt')


class MyMaze:

    def __init__(self, maze):
        self.colSize = 10
        self.rowSize = 10
        self.maze = maze

    recursionCount = 0

    def makeEntry(self, maze):
        num = random.randint(1, 8)
        maze[num][0] = 'E'
        return str(num) + '0'

    def checkBoundaries(self, x, y):
        if 0 <= x <= 9 or 0 <= y <= 9:
            return True
        else:
            return False

    def moveRight(self, row, col, maze, entry=False, check=False):
        if entry:
            return col + 1
        if row < 1 or row > 8 or col < 1 or row > 8:
            return col
        elif maze[row - 1][col] == 'P' and maze[row - 1][col + 1] == 'P':
            if check:
                return -1
            else:
                return col
        elif maze[row + 1][col] == 'P' and maze[row + 1][col + 1] == 'P':
            if check:
                return -1
            else:
                return col
        else:
            newCol = col + 1
            if newCol < 1 or newCol > 8:
                return -1
            elif maze[row - 1][newCol] == 'P' and maze[row - 1][newCol + 1] == 'P':
                return -1
            elif maze[row + 1][newCol] == 'P' and maze[row + 1][newCol + 1] == 'P':
                return -1
            else:
                return newCol

    def moveLeft(self, row, col, maze, check=False):
        if row < 1 or row > 8 or col < 1 or row > 8:
            return col
        elif maze[row - 1][col] == 'P' and maze[row - 1][col - 1] == 'P':
            if check:
                return -1
            else:
                return col
        elif maze[row + 1][col] == 'P' and maze[row + 1][col - 1] == 'P':
            if check:
                return -1
            else:
                return col
        else:
            newCol = col - 1
            if newCol < 1 or newCol > 8:
                return -1
            elif maze[row - 1][newCol] == 'P' and maze[row - 1][newCol - 1] == 'P':
                return -1
            elif maze[row + 1][newCol] == 'P' and maze[row + 1][newCol - 1] == 'P':
                return -1
            else:
                return newCol

    def moveUp(self, row, col, maze, check=False):
        if row < 1 or row > 8 or col < 1 or row > 8:
            return row
        elif maze[row][col - 1] == 'P' and maze[row - 1][col - 1] == 'P':
            if check:
                return -1
            else:
                return row
        elif maze[row][col + 1] == 'P' and maze[row - 1][col + 1] == 'P':
            if check:
                return -1
            else:
                return row
        else:
            newRow = row - 1
            if newRow < 1 or newRow > 8:
                return -1
            elif maze[newRow][col - 1] == 'P' and maze[newRow - 1][col - 1] == 'P':
                return -1
            elif maze[newRow][col + 1] == 'P' and maze[newRow - 1][col + 1] == 'P':
                return -1
            else:
                return newRow

    def moveDown(self, row, col, maze, check=False):
        if row < 1 or row > 8 or col < 1 or row > 8:
            return row
        elif maze[row][col - 1] == 'P' and maze[row + 1][col - 1] == 'P':
            if check:
                return -1
            else:
                return row
        elif maze[row][col + 1] == 'P' and maze[row + 1][col + 1] == 'P':
            if check:
                return -1
            else:
                return row
        else:
            newRow = row + 1
            if newRow < 1 or newRow > 8:
                return -1
            elif maze[newRow][col - 1] == 'P' and maze[newRow + 1][col - 1] == 'P':
                return -1
            elif maze[newRow][col + 1] == 'P' and maze[newRow + 1][col + 1] == 'P':
                return -1
            else:
                return newRow

    def checkSpaces(self, row, col, maze):
        if maze[row - 1][col] == 'P' and maze[row - 1][col + 1] == 'P':
            return False
        elif maze[row + 1][col] == 'P' and maze[row + 1][col + 1] == 'P':
            return False
        elif maze[row - 1][col] == 'P' and maze[row - 1][col - 1] == 'P':
            return False
        elif maze[row + 1][col] == 'P' and maze[row + 1][col - 1] == 'P':
            return False
        elif maze[row][col - 1] == 'P' and maze[row - 1][col - 1] == 'P':
            return False
        elif maze[row][col + 1] == 'P' and maze[row - 1][col + 1] == 'P':
            return False
        elif maze[row][col - 1] == 'P' and maze[row + 1][col - 1] == 'P':
            return False
        elif maze[row][col + 1] == 'P' and maze[row + 1][col + 1] == 'P':
            return False
        else:
            return True

    def exit(self, row, col, maze):
        if row == 1:
            maze[row - 1][col] = 'X'
        elif row == 8:
            maze[row + 1][col] = 'X'
        elif col == 8:
            maze[row][col + 1] = 'X'
        else:
            first = []
            second = []
            third = []

            for i in range(self.colSize):
                if maze[1][i] == 'P':
                    first.append((1, i))

            for i in range(self.colSize):
                if maze[8][i] == 'P':
                    second.append((8, i))

            for i in range(self.colSize):
                if maze[i][8] == 'P':
                    third.append((i, 8))

            done = False
            while not done:
                num = random.randint(1, 3)

                if num == 1:
                    '''first'''
                    if len(first) != 0:
                        done = True
                        anotherNum = random.randint(0, len(first) - 1)
                        coords = first[anotherNum]
                        maze[0][coords[1]] = 'X'
                elif num == 2:
                    '''two'''
                    if len(second) != 0:
                        done = True
                        anotherNum = random.randint(0, len(second) - 1)
                        coords = second[anotherNum]
                        maze[9][coords[1]] = 'X'
                elif num == 3:
                    '''three'''
                    if len(third) != 0:
                        done = True
                        anotherNum = random.randint(0, len(third) - 1)
                        coords = third[anotherNum]
                        maze[coords[0]][9] = 'X'

    def saveFile(self, maze):
        file = open(FILE, 'w')
        for i in range(self.rowSize):
            for j in range(self.colSize):
                file.write(maze[i][j] + '\n')
        file.close()
        # print('success')

    def getMazeCopy(self):
        return self.maze.copy()

    def getNextMove(self, row, col, maze):
        num = random.randint(1, 4)
        if self.moveRight(row, col, maze, False, True) == -1 and self.moveLeft(row, col, maze, True) == -1 and \
                self.moveUp(row, col, maze, True) == -1 and self.moveDown(row, col, maze, True) == -1:
            return row, col

        if self.recursionCount == 20:
            return -1, -1

        if num == 1:
            newCol = self.moveRight(row, col, maze)

            if newCol == -1:
                self.recursionCount += 1
                return self.getNextMove(row, col, maze)
            if not self.checkBoundaries(row, newCol):
                return row, newCol
            if maze[row][newCol] == 'P':
                self.recursionCount += 1
                return self.getNextMove(row, col, maze)
            else:
                return row, newCol
        elif num == 2:
            newCol = self.moveLeft(row, col, maze)

            if newCol == -1:
                self.recursionCount += 1
                return self.getNextMove(row, col, maze)
            if not self.checkBoundaries(row, newCol):
                return row, newCol
            if maze[row][newCol] == 'P':
                self.recursionCount += 1
                return self.getNextMove(row, col, maze)
            else:
                return row, newCol
        elif num == 3:
            newRow = self.moveUp(row, col, maze)

            if newRow == -1:
                self.recursionCount += 1
                return self.getNextMove(row, col, maze)
            if not self.checkBoundaries(row, newRow):
                return row, newRow
            if maze[newRow][col] == 'P':
                self.recursionCount += 1
                return self.getNextMove(row, col, maze)
            else:
                return newRow, col
        elif num == 4:
            newRow = self.moveDown(row, col, maze)

            if newRow == -1:
                self.recursionCount += 1
                return self.getNextMove(row, col, maze)
            if not self.checkBoundaries(row, newRow):
                return row, newRow
            if maze[newRow][col] == 'P':
                self.recursionCount += 1
                return self.getNextMove(row, col, maze)
            else:
                return newRow, col



from generator.subpackage.MazeGen import MyMaze


class Main:

    maze = ['W'] * 10

    for i in range(len(maze)):
        maze[i] = ['W'] * 10

    test = MyMaze(maze)

    path = []
    location = test.makeEntry(maze)
    row = int(location[0])
    col = int(location[1])

    colNum = test.moveRight(row, col, maze, True)

    location = str(row) + str(colNum)
    row = int(location[0])
    col = int(location[1])
    maze[row][col] = 'P'

    path.append((row, col))

    coords = test.getNextMove(row, col, maze)
    row = coords[0]
    col = coords[1]

    path.append(coords)

    previous = 0, 0
    attempts = 0

    while test.checkBoundaries(row, col):

        if test.checkBoundaries(row, col):
            try:
                if maze[row][col] != 'E' and maze[row][col] != 'P':
                    maze[row][col] = 'P'
            except IndexError:
                print('oops')

        # for blah in maze:
        #     print(' '.join([str(elem) for elem in blah]))

        if coords == previous:
            attempts += 1
        else:
            attempts = 0
        if attempts == 10:
            test.exit(row, col, maze)
            break
        previous = coords
        coords = test.getNextMove(row, col, maze)
        test.recursionCount = 0
        if coords[0] == -1 or coords[1] == -1:
            if len(path) > 0:
                coords = path.pop()
                row = coords[0]
                col = coords[1]
            else:
                break
        else:
            path.append(coords)
            row = coords[0]
            col = coords[1]

        # print(coords)

    # print()
    # for row in maze:
    #     print(' '.join([str(elem) for elem in row]))

    test.saveFile(maze)

    def getGenerator(self):
        return self.test



from subclass.Node import Node


class DepthSearch:

    def __init__(self, maze=None):

        if maze is None:
            maze = []
        self.file = open("resources/maze1.txt", 'r').readlines()
        self.stack = []
        self.rowSize = 10
        self.colSize = 10
        self.completedNodes = []
        self.completePath = []

        self.mazeArray = maze
        if len(self.mazeArray) == 0:
            self.mazeArray = ['O'] * self.colSize
            self.completedNodes = []
            self.completePath = []

            for i in range(self.rowSize):
                self.mazeArray[i] = ['O'] * self.colSize

            index = 0
            for i in range(self.rowSize):
                for j in range(self.colSize):
                    self.mazeArray[i][j] = self.file[index].strip()
                    index += 1

    def createNode(self, location, path):
        node = {
            'location': location,
            'data': 'P',
            'path': path
        }
        return node

    def findEntry(self):
        for i in range(self.colSize):
            for j in range(self.rowSize):
                if self.mazeArray[i][0] == 'E':
                    return str(i) + '0'

    def isGoal(self, node):
        if node.data == 'X':
            return True
        else:
            return False

    def checkForChildren(self, node):
        children = []
        location = node.location
        row = int(location[0])
        col = int(location[1])

        if (self.mazeArray[row + 1][col] == 'P' or self.mazeArray[row + 1][col] == 'X') and row != 0:
            children.append(str(row + 1) + str(col))

        if (self.mazeArray[row][col - 1] == 'P' or self.mazeArray[row][col - 1] == 'X') and col != 0:
            children.append(str(row) + str(col - 1))

        if (self.mazeArray[row - 1][col] == 'P' or self.mazeArray[row - 1][col] == 'X') and row != 9:
            children.append(str(row - 1) + str(col))

        if (self.mazeArray[row][col + 1] == 'P' or self.mazeArray[row][col + 1] == 'X') and col != 9:
            children.append(str(row) + str(col + 1))

        return children

    def getData(self, id):
        row = id[0]
        col = id[1]
        return self.mazeArray[int(row)][int(col)]

    def checkVisited(self, node):
        if node.location in self.completedNodes:
            return True
        else:
            return False

    def underline(self, list, path):
        for i in range(self.colSize):
            for j in range(self.rowSize):
                loc = str(i) + str(j)
                if loc in path:
                    letter = list[i][j]
                    list[i][j] = '*'

    def run(self):
        nodeid = self.findEntry()
        self.completedNodes.append(nodeid)
        self.completePath.append(nodeid)
        node = Node(nodeid, self.getData(nodeid), self.completedNodes)
        self.stack.append(node)

        if not self.isGoal(self.stack.pop()):
            children = self.checkForChildren(node)
            for child in children:
                if child not in self.completedNodes:
                    self.completedNodes.append(child)
                    newNode = Node(child, self.getData(child), self.completePath, node)
                    self.stack.append(newNode)

        try:
            newNode
        except NameError:
            currentNode = node
        else:
            currentNode = self.stack.pop()
            self.completePath.append(currentNode.location)

        while not self.isGoal(currentNode):

            children = self.checkForChildren(currentNode)
            index = 0
            for child in children:
                if child not in self.completedNodes:
                    self.completedNodes.append(child)
                    newNode = Node(child, self.getData(child), self.completePath, currentNode)
                    self.stack.append(newNode)
                else:
                    try:
                        if len(children) == 1:
                            currentNode = currentNode.parent
                    except ValueError:
                        '''do nothing'''
                index += 1

            try:
                currentNode = self.stack.pop()
                self.completePath.append(currentNode.location)
            except IndexError:
                print('There is not a valid path.')
                return

            # for node in currentNode.path:
            #     print(node.location)

        finalPath = []

        for myNode in currentNode.path:
            finalPath.append(myNode.location)

        self.underline(self.mazeArray, finalPath)

        for row in self.mazeArray:
            print(' '.join([str(elem) for elem in row]))

        print()
        print(finalPath)



from anytree import NodeMixin


class Node(NodeMixin):
    location = ''
    data = ''
    traveledPath = []

    def __init__(self, location, data, path, parent=None):
        super(Node, self).__init__()
        self.location = location
        self.data = data
        self.traveledPath = path
        self.parent = parent
